import json
import os
from tempfile import mkdtemp


from regipy import NoRegistrySubkeysException
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.utils import dump_hive_to_json

from regipy.recovery import apply_transaction_logs
from regipy.regdiff import compare_hives
from regipy.registry import RegistryHive, NKRecord
from regipy.cli_utils import get_filtered_subkeys


def test_parse_header(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)

    assert isinstance(registry_hive, RegistryHive)
    assert registry_hive.header.primary_sequence_num == 749
    assert registry_hive.header.secondary_sequence_num == 749
    assert registry_hive.header.last_modification_time == 129782982453388850
    assert registry_hive.header.major_version == 1
    assert registry_hive.header.minor_version == 3
    assert registry_hive.header.root_key_offset == 32
    assert registry_hive.header.hive_bins_data_size == 733184
    assert registry_hive.header.minor_version == 3
    assert registry_hive.header.file_name == "?\\C:\\Users\\vibranium\\ntuser.dat"
    assert registry_hive.header.checksum == 476614345


def test_parse_root_key(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)

    assert isinstance(registry_hive, RegistryHive)
    assert isinstance(registry_hive.root, NKRecord)
    assert (
        registry_hive.root.name
        == "CMI-CreateHive{6A1C4018-979D-4291-A7DC-7AED1C75B67C}"
    )
    assert registry_hive.root.subkey_count == 11
    assert dict(registry_hive.root.header) == {
        "access_bits": b"\x02\x00\x00\x00",
        "class_name_offset": 4294967295,
        "class_name_size": 0,
        "flags": {
            "KEY_COMP_NAME": True,
            "KEY_HIVE_ENTRY": True,
            "KEY_HIVE_EXIT": False,
            "KEY_NO_DELETE": True,
            "KEY_PREDEF_HANDLE": False,
            "KEY_SYM_LINK": False,
            "KEY_VOLATILE": False,
        },
        "key_name_size": 52,
        "key_name_string": b"CMI-CreateHive{6A1C4018-979D-4291-A7DC-7AED1C75B67C}",
        "largest_sk_class_name": 0,
        "largest_sk_name": 40,
        "largest_value_name": 0,
        "last_modified": 129780243434537497,
        "largest_value_data": 0,
        "parent_key_offset": 1968,
        "security_key_offset": 1376,
        "subkey_count": 11,
        "subkeys_list_offset": 73760,
        "values_count": 0,
        "values_list_offset": 4294967295,
        "volatile_subkey_count": 0,
        "volatile_subkeys_list_offset": 4294967295,
    }


def test_find_keys_ntuser(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)
    run_key = registry_hive.get_key(r"\Software\Microsoft\Windows\CurrentVersion\Run")

    assert run_key.name == "Run"
    assert run_key.header.last_modified == 129779615948377168

    values = [x for x in run_key.iter_values(as_json=True)]
    assert values[0].name == "Sidebar"
    assert values[0].value_type == "REG_EXPAND_SZ"


def test_find_keys_partial_ntuser_hive(ntuser_software_partial):
    registry_hive = RegistryHive(
        ntuser_software_partial,
        hive_type=NTUSER_HIVE_TYPE,
        partial_hive_path=r"\Software",
    )

    run_key = registry_hive.get_key(r"\Software\Microsoft\Windows\CurrentVersion\Run")
    assert run_key.name == "Run"
    assert run_key.header.last_modified == 132024690510209250

    values = [x for x in run_key.iter_values(as_json=True)]
    assert values[0].name == "OneDrive"
    assert values[0].value_type == "REG_SZ"


def test_regdiff(ntuser_hive, second_hive_path):
    found_differences = compare_hives(ntuser_hive, second_hive_path, verbose=True)
    assert len(found_differences) == 7
    assert len([x for x in found_differences if x[0] == "new_subkey"]) == 6
    assert len([x for x in found_differences if x[0] == "new_value"]) == 1


def test_ntuser_emojis(transaction_ntuser):
    # There are some cases where the Registry stores utf-16 emojis as subkey names :)
    registry_hive = RegistryHive(transaction_ntuser)
    international = registry_hive.get_key(r"\Control Panel\International")
    subkeys = [x.name for x in international.iter_subkeys()]
    assert subkeys == ["Geo", "User Profile", "User Profile System Backup", "🌎🌏🌍"]


def test_recurse_ntuser(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)

    value_types = {
        "REG_BINARY": 0,
        "REG_DWORD": 0,
        "REG_EXPAND_SZ": 0,
        "REG_MULTI_SZ": 0,
        "REG_NONE": 0,
        "REG_QWORD": 0,
        "REG_SZ": 0,
    }

    subkey_count = 0
    values_count = 0
    for subkey in registry_hive.recurse_subkeys(as_json=True):
        subkey_values = subkey.values
        subkey_count += 1
        values_count += len(subkey_values or [])
        if subkey_values:
            for x in subkey_values:
                value_types[x["value_type"]] += 1

    assert subkey_count == 1812
    assert values_count == 4094
    assert value_types == {
        "REG_BINARY": 531,
        "REG_DWORD": 1336,
        "REG_EXPAND_SZ": 93,
        "REG_MULTI_SZ": 303,
        "REG_NONE": 141,
        "REG_QWORD": 54,
        "REG_SZ": 1636,
    }


def test_recurse_partial_ntuser(ntuser_software_partial):
    registry_hive = RegistryHive(
        ntuser_software_partial,
        hive_type=NTUSER_HIVE_TYPE,
        partial_hive_path=r"\Software",
    )
    for subkey_count, subkey in enumerate(registry_hive.recurse_subkeys(as_json=True)):
        assert subkey.actual_path.startswith(registry_hive.partial_hive_path)
    assert subkey_count == 6395


def test_recurse_ntuser_without_fetching_values(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)
    for subkey_count, subkey in enumerate(
        registry_hive.recurse_subkeys(as_json=True, fetch_values=False)
    ):
        assert subkey.values == []
        assert subkey.values_count >= 0
    assert subkey_count == 1811


def test_recurse_amcache(amcache_hive):
    registry_hive = RegistryHive(amcache_hive)

    value_types = {
        "REG_BINARY": 0,
        "REG_DWORD": 0,
        "REG_EXPAND_SZ": 0,
        "REG_MULTI_SZ": 0,
        "REG_NONE": 0,
        "REG_QWORD": 0,
        "REG_SZ": 0,
    }
    subkey_count = 0
    values_count = 0
    for subkey in registry_hive.recurse_subkeys():
        subkey_count += 1
        subkey_values = subkey.values
        values_count += len(subkey_values or [])
        if subkey_values:
            for x in subkey_values:
                value_types[x.value_type] += 1
    assert subkey_count == 2105
    assert values_count == 17539
    assert value_types == {
        "REG_BINARY": 56,
        "REG_DWORD": 1656,
        "REG_EXPAND_SZ": 0,
        "REG_MULTI_SZ": 140,
        "REG_NONE": 0,
        "REG_QWORD": 1254,
        "REG_SZ": 14433,
    }


def test_ntuser_apply_transaction_logs(transaction_ntuser, transaction_log):
    output_path = os.path.join(mkdtemp(), "recovered_hive.dat")
    restored_hive_path, recovered_dirty_pages_count = apply_transaction_logs(
        transaction_ntuser, transaction_log, restored_hive_path=output_path
    )
    assert recovered_dirty_pages_count == 132

    found_differences = compare_hives(transaction_ntuser, restored_hive_path)
    assert len(found_differences) == 588
    assert len([x for x in found_differences if x[0] == "new_subkey"]) == 527
    assert len([x for x in found_differences if x[0] == "new_value"]) == 60


def test_system_apply_transaction_logs(
    transaction_system, system_tr_log_1, system_tr_log_2
):
    output_path = os.path.join(mkdtemp(), "recovered_hive.dat")
    restored_hive_path, recovered_dirty_pages_count = apply_transaction_logs(
        transaction_system,
        primary_log_path=system_tr_log_1,
        secondary_log_path=system_tr_log_2,
        restored_hive_path=output_path,
    )
    assert recovered_dirty_pages_count == 315

    found_differences = compare_hives(transaction_system, restored_hive_path)
    assert len(found_differences) == 2511
    assert len([x for x in found_differences if x[0] == "new_subkey"]) == 2458
    assert len([x for x in found_differences if x[0] == "new_value"]) == 53


def test_system_hive_devprop_structure(system_devprop):
    registry_hive = RegistryHive(system_devprop)
    subkey = registry_hive.get_key(
        "\\ControlSet001\\Enum\\ACPI\\ACPI0003\\0\\Properties\\{83da6326-97a6-4088-9453-"
        "a1923f573b29}\\0003"
    )
    assert subkey.values_count == 1
    value = subkey.get_values()[0]
    assert value.name == "(default)"
    assert (
        value.value
        == "cmbatt.inf:db04a16c09a7808a:AcAdapter_Inst:6.3.9600.16384:ACPI\\ACPI0003"
    )
    assert value.value_type == 18


def test_system_apply_transaction_logs_2(
    transaction_usrclass, usrclass_tr_log_1, usrclass_tr_log_2
):
    output_path = os.path.join(mkdtemp(), "recovered_hive.dat")
    restored_hive_path, recovered_dirty_pages_count = apply_transaction_logs(
        transaction_usrclass,
        primary_log_path=usrclass_tr_log_1,
        secondary_log_path=usrclass_tr_log_2,
        restored_hive_path=output_path,
    )
    assert recovered_dirty_pages_count == 158

    found_differences = compare_hives(transaction_usrclass, restored_hive_path)
    assert len(found_differences) == 225
    assert len([x for x in found_differences if x[0] == "new_subkey"]) == 93
    assert len([x for x in found_differences if x[0] == "new_value"]) == 132


def test_hive_serialization(ntuser_hive, temp_output_file):
    registry_hive = RegistryHive(ntuser_hive)
    dump_hive_to_json(
        registry_hive, temp_output_file, registry_hive.root, verbose=False
    )
    counter = 0
    with open(temp_output_file, "r") as dumped_hive:
        for x in dumped_hive.readlines():
            assert json.loads(x)
            counter += 1
    assert counter == 1812


def test_get_key(software_hive):
    """
    # Refers to https://github.com/mkorman90/regipy/issues/144
    """
    registry_hive = RegistryHive(software_hive)
    # We verify the registry headers are similar, because this is the same subkey.
    assert (
        registry_hive.get_key("ODBC").header
        == registry_hive.root.get_subkey("ODBC").header
    )
    assert (
        registry_hive.root.get_subkey("ODBC").header
        == registry_hive.get_key("SOFTWARE\\ODBC").header
    )


def test_get_subkey_errors(software_hive):
    registry_hive = RegistryHive(software_hive)
    # Tests the NoRegistrySubkeysException that suppose to be raised
    try:
        registry_hive.get_key("ODBC").get_subkey("xyz")
        assert False
    except NoRegistrySubkeysException:
        assert True

    # Tests value if raised_on_missing is set to False
    assert (
        registry_hive.get_key("ODBC").get_subkey("xyz", raise_on_missing=False) is None
    )


def test_parse_security_info(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)
    run_key = registry_hive.get_key(r"\Software\Microsoft\Windows\CurrentVersion\Run")

    security_key_info = run_key.get_security_key_info()
    assert security_key_info["owner"] == "S-1-5-18"
    assert security_key_info["group"] == "S-1-5-18"
    assert len(security_key_info["dacl"]) == 4
    assert security_key_info["dacl"][0] == {
        "access_mask": {
            "ACCESS_SYSTEM_SECURITY": False,
            "DELETE": True,
            "GENERIC_ALL": False,
            "GENERIC_EXECUTE": False,
            "GENERIC_READ": False,
            "GENERIC_WRITE": False,
            "MAXIMUM_ALLOWED": False,
            "READ_CONTROL": True,
            "SYNCHRONIZE": False,
            "WRITE_DAC": True,
            "WRITE_OWNER": True,
        },
        "ace_type": "ACCESS_ALLOWED",
        "flags": {
            "CONTAINER_INHERIT_ACE": True,
            "INHERIT_ONLY_ACE": False,
            "NO_PROPAGATE_INHERIT_ACE": False,
            "OBJECT_INHERIT_ACE": True,
        },
        "sid": "S-1-5-21-2036804247-3058324640-2116585241-1673",
    }

    dacl_sids = [x["sid"] for x in security_key_info["dacl"]]
    assert dacl_sids == [
        "S-1-5-21-2036804247-3058324640-2116585241-1673",
        "S-1-5-18",
        "S-1-5-32-544",
        "S-1-5-12",
    ]


def test_parse_filetime_value(system_hive_with_filetime):
    registry_hive = RegistryHive(system_hive_with_filetime)
    subkey = registry_hive.get_key(
        r"\ControlSet001\Enum\USBSTOR\Disk&Ven_SanDisk&Prod_Cruzer&Rev_1.2"
        r"0\200608767007B7C08A6A&0\Properties\{83da6326-97a6-4088-9453-a1923f573b29}\0064"
    )
    val = subkey.get_value("(default)", as_json=True)
    assert val == "2020-03-17T14:02:38.955490+00:00"


def test_ntuser_filtered_timestamps_do_not_fetch_values(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)
    for subkey_count, entry in enumerate(
        get_filtered_subkeys(
            registry_hive,
            registry_hive.root,
            fetch_values=False,
            start_date="2012-04-03T00:00:00.000000",
            end_date="2012-04-03T23:59:59.999999",
        )
    ):

        assert entry.values == []
    assert subkey_count == 1489


def test_ntuser_filtered_timestamps_fetch_values(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)
    for subkey_count, entry in enumerate(
        get_filtered_subkeys(
            registry_hive,
            registry_hive.root,
            fetch_values=True,
            start_date="2012-04-03T00:00:00.000000",
            end_date="2012-04-03T23:59:59.999999",
        )
    ):

        # values_count is parsed from the subkey header, so this test if effective.
        if entry.values_count > 0:
            assert len(entry.values) == entry.values_count
    assert subkey_count == 1489


def test_ntuser_filtered_timestamps_no_filter(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)
    for subkey_count, entry in enumerate(
        get_filtered_subkeys(registry_hive, registry_hive.root, fetch_values=False)
    ):
        assert entry.values == []
    assert subkey_count == 1811



import logging

from regipy import RegistryKeyNotFoundException, convert_wintime
from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)


SAFEBOOT_NETWORK_PATH = r"Control\SafeBoot\Network"
SAFEBOOT_MINIMAL_PATH = r"Control\SafeBoot\Minimal"


class SafeBootConfigurationPlugin(Plugin):
    NAME = "safeboot_configuration"
    DESCRIPTION = "Get safeboot configuration"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def _get_safeboot_entries(self, subkey_path):
        entries = []
        for safeboot_network_path in self.registry_hive.get_control_sets(subkey_path):
            control_set_name = safeboot_network_path.split("\\")[1]
            try:
                subkey = self.registry_hive.get_key(safeboot_network_path)
            except RegistryKeyNotFoundException as ex:
                logger.error(ex)
                continue

            for safeboot_network_subkey in subkey.iter_subkeys():
                timestamp = convert_wintime(
                    safeboot_network_subkey.header.last_modified, as_json=self.as_json
                )
                entries.append(
                    {
                        "timestamp": timestamp,
                        "name": safeboot_network_subkey.name,
                        "description": safeboot_network_subkey.get_value(),
                        "control_set_name": control_set_name,
                    }
                )
        return entries

    def run(self):
        logger.debug("Started Safeboot Configuration Plugin...")

        self.entries = {
            "network": self._get_safeboot_entries(SAFEBOOT_NETWORK_PATH),
            "minimal": self._get_safeboot_entries(SAFEBOOT_MINIMAL_PATH),
        }



import logging

from regipy.exceptions import NoRegistrySubkeysException, RegistryKeyNotFoundException
from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

USBSTOR_KEY_PATH = r"Enum\USBSTOR"
DISK_GUID_PATH = r"Device Parameters\Partmgr"
PROPERTIES_NAME_GUID = r"{540b947e-8b40-45bc-a8a2-6a0b894cbda2}"
PROPERTIES_DATES_GUID = r"{83da6326-97a6-4088-9453-a1923f573b29}"
DEVICE_NAME_KEY = "0004"
FIRST_INSTALLED_TIME_KEY = "0065"
LAST_CONNECTED_TIME_KEY = "0066"
LAST_REMOVED_TIME_KEY = "0067"
LAST_INSTALLED_TIME_KEY = "0064"


class USBSTORPlugin(Plugin):
    NAME = "usbstor_plugin"
    DESCRIPTION = "Parse the connected USB devices history"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        try:
            for subkey_path in self.registry_hive.get_control_sets(USBSTOR_KEY_PATH):
                usbstor_key = self.registry_hive.get_key(subkey_path)
                for usbstor_drive in usbstor_key.iter_subkeys():
                    try:
                        disk, manufacturer, title, version = usbstor_drive.name.split(
                            "&"
                        )
                    except ValueError:
                        manufacturer, title, version = None, None, None

                    for serial_subkey in usbstor_drive.iter_subkeys():
                        timestamp = convert_wintime(
                            serial_subkey.header.last_modified, as_json=self.as_json
                        )
                        serial_number = serial_subkey.name
                        key_path = (
                            rf"{subkey_path}\{usbstor_drive.name}\{serial_number}"
                        )

                        try:
                            device_guid_key = self.registry_hive.get_key(
                                rf"{subkey_path}\{usbstor_drive.name}\{serial_number}\{DISK_GUID_PATH}"
                            )
                            disk_guid = device_guid_key.get_value("DiskId")
                        except RegistryKeyNotFoundException:
                            disk_guid = None

                        properties_subkey = serial_subkey.get_subkey("Properties")
                        if not properties_subkey:
                            (
                                device_name,
                                first_installed_time,
                                last_connected_time,
                                last_removed_time,
                                last_installed_time,
                            ) = (None, None, None, None, None)
                        else:
                            try:
                                device_name_guid_key = properties_subkey.get_subkey(
                                    PROPERTIES_NAME_GUID
                                )
                                device_name_key = device_name_guid_key.get_subkey(
                                    DEVICE_NAME_KEY
                                )
                                device_name = device_name_key.get_value()
                            except (
                                RegistryKeyNotFoundException,
                                NoRegistrySubkeysException,
                            ):
                                device_name = None

                            (
                                first_installed_time,
                                last_connected_time,
                                last_removed_time,
                                last_installed_time,
                            ) = (None, None, None, None)

                            dates_subkey = properties_subkey.get_subkey(
                                PROPERTIES_DATES_GUID, raise_on_missing=False
                            )
                            if dates_subkey:
                                first_installed_key = dates_subkey.get_subkey(
                                    FIRST_INSTALLED_TIME_KEY, raise_on_missing=False
                                )
                                if first_installed_key:
                                    first_installed_time = (
                                        first_installed_key.get_value(
                                            as_json=self.as_json
                                        )
                                    )

                                last_connected_key = dates_subkey.get_subkey(
                                    LAST_CONNECTED_TIME_KEY, raise_on_missing=False
                                )
                                if last_connected_key:
                                    last_connected_time = last_connected_key.get_value(
                                        as_json=self.as_json
                                    )

                                last_removed_key = dates_subkey.get_subkey(
                                    LAST_REMOVED_TIME_KEY, raise_on_missing=False
                                )
                                if last_removed_key:
                                    last_removed_time = last_removed_key.get_value(
                                        as_json=self.as_json
                                    )

                                last_installed_key = dates_subkey.get_subkey(
                                    LAST_INSTALLED_TIME_KEY, raise_on_missing=False
                                )
                                if last_installed_key:
                                    last_installed_time = last_installed_key.get_value(
                                        as_json=self.as_json
                                    )

                        self.entries.append(
                            {
                                "last_write": timestamp,
                                "key_path": key_path,
                                "last_connected": last_connected_time,
                                "last_removed": last_removed_time,
                                "first_installed": first_installed_time,
                                "last_installed": last_installed_time,
                                "serial_number": serial_number,
                                "device_name": device_name,
                                "disk_guid": disk_guid,
                                "manufacturer": manufacturer,
                                "version": version,
                                "title": title,
                            }
                        )

        except RegistryKeyNotFoundException as ex:
            logger.error(
                f"Could not find {self.NAME} plugin data at: {USBSTOR_KEY_PATH}: {ex}"
            )



import logging

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import get_subkey_values_from_list

logger = logging.getLogger(__name__)


ROUTES_PATH = r"Services\Tcpip\Parameters\PersistentRoutes"


class RoutesPlugin(Plugin):
    NAME = "routes"
    DESCRIPTION = "Get list of routes"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        logger.debug("Started Routes Plugin...")

        routes_path_list = self.registry_hive.get_control_sets(ROUTES_PATH)
        self.entries = get_subkey_values_from_list(
            self.registry_hive, routes_path_list, as_json=self.as_json
        )



import logging
import struct

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime
from regipy.utils import convert_filetime2
from regipy.exceptions import RegistryKeyNotFoundException


logger = logging.getLogger(__name__)


DIAGSR_PATH = r"Services\VSS\Diag\SystemRestore"
crash_items = ("CrashDumpEnabled", "DumpFile", "MinidumpDir", "LogEvent")


class DiagSRPlugin(Plugin):
    NAME = "diag_sr"
    DESCRIPTION = "Get Diag\\SystemRestore values and data"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def can_run(self):
        return self.registry_hive.hive_type == SYSTEM_HIVE_TYPE

    def run(self):
        self.entries = {}
        diagsr_subkeys = self.registry_hive.get_control_sets(DIAGSR_PATH)
        for diagsr_subkey in diagsr_subkeys:
            try:
                diagsr = self.registry_hive.get_key(diagsr_subkey)
            except RegistryKeyNotFoundException as ex:
                logger.error(
                    f"Could not find {self.NAME} subkey at {diagsr_subkey}: {ex}"
                )
                continue
            self.entries[diagsr_subkey] = {
                "last_write": convert_wintime(diagsr.header.last_modified).isoformat()
            }
            for val in diagsr.iter_values():
                self.entries[diagsr_subkey][val.name] = convert_filetime2(
                    val.value[16:32]
                )



import logging

from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

NETWORK_DRIVES = r"\Network"


class NetworkDrivesPlugin(Plugin):
    NAME = "network_drives_plugin"
    DESCRIPTION = "Parse the user's mapped network drives"
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def run(self):
        try:
            network_drives = self.registry_hive.get_key(NETWORK_DRIVES)
            for mapped_drive in network_drives.iter_subkeys():
                timestamp = convert_wintime(
                    mapped_drive.header.last_modified, as_json=self.as_json
                )
                self.entries.append(
                    {
                        "last_write": timestamp,
                        "drive_letter": mapped_drive.name,
                        "network_path": mapped_drive.get_value("RemotePath"),
                    }
                )

        except RegistryKeyNotFoundException as ex:
            logger.error(
                f"Could not find {self.NAME} plugin data at: {NETWORK_DRIVES}: {ex}"
            )



from typing import Any
from regipy.structs import ACL, ACE, SID
from construct import Int64ub


def convert_sid(sid: Any, strip_rid: bool = False) -> str:
    identifier_authority = Int64ub.parse(b"\x00\x00" + sid.identifier_authority)
    sub_authorities = sid.subauthority[:-1] if strip_rid else sid.subauthority
    sub_identifier_authorities = "-".join(str(x) for x in sub_authorities)
    return f"S-{sid.revision}-{identifier_authority}-{sub_identifier_authorities}"


def get_acls(s):
    aces = []
    dacl = ACL.parse_stream(s)
    for _ in range(dacl.ace_count):
        parsed_ace = ACE.parse_stream(s)
        ace_sid = SID.parse(parsed_ace.sid)

        aces.append(
            {
                "ace_type": str(parsed_ace.type),
                "flags": dict(parsed_ace.flags),
                "access_mask": dict(parsed_ace.access_mask),
                "sid": convert_sid(ace_sid),
            }
        )
    return aces



import lzma
import os

from pathlib import Path
from tempfile import mktemp

import pytest


def extract_lzma(path):
    tempfile_path = mktemp()
    with open(tempfile_path, "wb") as tmp:
        with lzma.open(path) as f:
            tmp.write(f.read())
    return tempfile_path


@pytest.fixture()
def temp_output_file():
    tempfile_path = mktemp()
    yield tempfile_path
    os.remove(tempfile_path)


@pytest.fixture(scope="module")
def test_data_dir():
    return str(Path(__file__).parent.joinpath("data"))


@pytest.fixture(scope="module")
def ntuser_hive(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "NTUSER.DAT.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def software_hive(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "SOFTWARE.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def system_hive(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "SYSTEM.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def sam_hive(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "SAM.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def security_hive(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "SECURITY.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def amcache_hive(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "amcache.hve.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def bcd_hive(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "BCD.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def second_hive_path(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "NTUSER_modified.DAT.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def transaction_ntuser(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "transactions_NTUSER.DAT.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def transaction_log(test_data_dir):
    temp_path = extract_lzma(
        os.path.join(test_data_dir, "transactions_ntuser.dat.log1.xz")
    )
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def transaction_system(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "SYSTEM_B.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def system_tr_log_1(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "SYSTEM_B.LOG1.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def system_tr_log_2(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "SYSTEM_B.LOG2.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def transaction_usrclass(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "UsrClass.dat.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def usrclass_tr_log_1(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "UsrClass.dat.LOG1.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def usrclass_tr_log_2(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "UsrClass.dat.LOG2.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def ntuser_software_partial(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "ntuser_software_partial.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def corrupted_system_hive(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "corrupted_system_hive.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def system_devprop(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "SYSTEM_2.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def ntuser_hive_2(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "NTUSER_with_winscp.DAT.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def shellbags_ntuser(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "NTUSER_BAGMRU.DAT.xz"))
    yield temp_path
    os.remove(temp_path)


@pytest.fixture(scope="module")
def system_hive_with_filetime(test_data_dir):
    temp_path = extract_lzma(os.path.join(test_data_dir, "SYSTEM_WIN_10_1709.xz"))
    yield temp_path
    os.remove(temp_path)



import logging

from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

# Ressources : https://patrickwu.space/2020/07/19/wsl-related-registry/

WSL_PATH = r"\Software\Microsoft\Windows\CurrentVersion\Lxss"


class WSLPlugin(Plugin):
    NAME = "wsl"
    DESCRIPTION = "Get WSL information"
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def get_wls_info(self, subkey, distribs=None):
        if distribs is None:
            distribs = []

        try:
            flags = subkey.get_value("Flags")
            state = subkey.get_value("State")
            version = subkey.get_value("Version")

            # Initialize the entry for a distribution with its GUID as the key
            distribution_entry = {
                "GUID": subkey.name,
                "last_modified": convert_wintime(
                    subkey.header.last_modified, as_json=self.as_json
                ),
                "wsl_distribution_source_location": subkey.get_value("BasePath"),
                "default_uid": subkey.get_value("DefaultUid"),
                "distribution_name": subkey.get_value("DistributionName"),
                "default_environment": subkey.get_value(
                    "DefaultEnvironment"
                ),  # REG_MULTI_SZ
                "flags": flags,
                "kernel_command_line": subkey.get_value("KernelCommandLine"),
                "package_family_name": subkey.get_value("PackageFamilyName"),
                "state": state,
                "filesystem": (
                    "lxfs" if version == 1 else "wslfs" if version == 2 else "Unknown"
                ),
            }

            # Decode flags for additional information
            if flags is not None:
                distribution_entry["enable_interop"] = bool(flags & 0x1)
                distribution_entry["append_nt_path"] = bool(flags & 0x2)
                distribution_entry["enable_drive_mounting"] = bool(flags & 0x4)

            # Decode the state of the distribution
            if state is not None:
                if state == 0x1:
                    distribution_entry["state"] = "Normal"
                elif state == 0x3:
                    distribution_entry["state"] = "Installing"
                elif state == 0x4:
                    distribution_entry["state"] = "Uninstalling"
                else:
                    distribution_entry["state"] = "Unknown"

            # Add the distribution entry with its GUID to the list of distributions
            distribs.append(distribution_entry)

        except Exception as e:
            logger.error(f"Error processing subkey {subkey.name}: {e}")
            raise

        return distribs

    def run(self):

        try:
            # Attempt to get the WSL registry key
            wsl_key = self.registry_hive.get_key(WSL_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(f"Registry key not found at path {WSL_PATH}: {ex}")
            return

        self.entries = {
            WSL_PATH: {
                "last_modified": convert_wintime(
                    wsl_key.header.last_modified, as_json=self.as_json
                ),
                "number_of_distrib": wsl_key.header.subkey_count,
                "default_distrib_GUID": wsl_key.get_value("DefaultDistribution"),
                "wsl_version": (
                    "WSL1"
                    if wsl_key.get_value("DefaultVersion") == 1
                    else (
                        "WSL2"
                        if wsl_key.get_value("DefaultVersion") == 2
                        else "Unknown"
                    )
                ),
                "nat_ip_address": wsl_key.get_value("NatIpAddress"),
                "distributions": [],
            }
        }

        try:
            for distrib in wsl_key.iter_subkeys():
                distribs = self.get_wls_info(distrib)
                self.entries[WSL_PATH]["distributions"] = distribs
        except Exception as e:
            logger.error(f"Error iterating over subkeys in {distrib.path}: {e}")



import logging

from construct import Int64ul

from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

BAM_PATH = [r"Services\bam\UserSettings", r"Services\bam\state\UserSettings"]


class BAMPlugin(Plugin):
    NAME = "background_activity_moderator"
    DESCRIPTION = "Get the computer name"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        logger.debug("Started Computer Name Plugin...")

        for path in BAM_PATH:
            try:
                for subkey_path in self.registry_hive.get_control_sets(path):
                    subkey = self.registry_hive.get_key(subkey_path)
                    for sid_subkey in subkey.iter_subkeys():

                        sid = sid_subkey.name
                        logger.debug(f"Parsing BAM for {sid}")
                        sequence_number = None
                        version = None
                        entries = []

                        for value in sid_subkey.get_values(
                            trim_values=self.trim_values
                        ):
                            if value.name == "SequenceNumber":
                                sequence_number = value.value
                            elif value.name == "Version":
                                version = value.value
                            else:
                                entries.append(
                                    {
                                        "executable": value.name,
                                        "timestamp": convert_wintime(
                                            Int64ul.parse(value.value),
                                            as_json=self.as_json,
                                        ),
                                    }
                                )

                        self.entries.extend(
                            [
                                {
                                    "sequence_number": sequence_number,
                                    "version": version,
                                    "sid": sid,
                                    "key_path": f"{subkey_path}\\{sid}",
                                    **x,
                                }
                                for x in entries
                            ]
                        )
            except RegistryKeyNotFoundException as ex:
                logger.error(ex)



import logging

from regipy.exceptions import RegistryValueNotFoundException
from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

COMPUTER_NAME_PATH = r"Control\ComputerName\ComputerName"


class ComputerNamePlugin(Plugin):
    NAME: str = "computer_name"
    DESCRIPTION = "Get the computer name"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        logger.debug("Started Computer Name Plugin...")

        for subkey_path in self.registry_hive.get_control_sets(COMPUTER_NAME_PATH):
            subkey = self.registry_hive.get_key(subkey_path)

            try:
                self.entries.append(
                    {
                        "name": subkey.get_value("ComputerName", as_json=self.as_json),
                        "timestamp": convert_wintime(
                            subkey.header.last_modified, as_json=self.as_json
                        ),
                    }
                )
            except RegistryValueNotFoundException:
                logger.exception("Could not get computer name")
                continue



# Andrew Davis, andrew.davis@mandiant.com
# Copyright 2012 Mandiant
#
# Mandiant licenses this file to you under the Apache License, Version
# 2.0 (the "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at:
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.
#
# Identifies and parses Application Compatibility Shim Cache entries for forensic data.


# Original Mandiant shim cache parser, ported to Python 3
# Identifies and parses Application Compatibility Shim Cache entries for forensic data.

import datetime
import io as sio
import struct

import logging
import pytz

CACHE_MAGIC_NT5_2 = 0xBADC0FFE
CACHE_HEADER_SIZE_NT5_2 = 0x8
NT5_2_ENTRY_SIZE32 = 0x18
NT5_2_ENTRY_SIZE64 = 0x20

# Values used by Windows 6.1 (Win7 and Server 2008 R2)
CACHE_MAGIC_NT6_1 = 0xBADC0FEE
CACHE_HEADER_SIZE_NT6_1 = 0x80
NT6_1_ENTRY_SIZE32 = 0x20
NT6_1_ENTRY_SIZE64 = 0x30
CSRSS_FLAG = 0x2

# Values used by Windows 5.1 (WinXP 32-bit)
WINXP_MAGIC32 = 0xDEADBEEF
WINXP_HEADER_SIZE32 = 0x190
WINXP_ENTRY_SIZE32 = 0x228
MAX_PATH = 520

# Values used by Windows 8
WIN8_STATS_SIZE = 0x80
WIN8_MAGIC = b"00ts"

# Magic value used by Windows 8.1
WIN81_MAGIC = b"10ts"

# Values used by Windows 10
WIN10_STATS_SIZE = 0x30
WIN10_MAGIC = b"10ts"
CACHE_HEADER_SIZE_NT6_4 = 0x30
CACHE_MAGIC_NT6_4 = 0x30

BAD_ENTRY_DATA = "N/A"
G_VERBOSE = False
G_USE_BOM = False
OUTPUT_HEADER = ["Last Modified", "Last Update", "Path", "File Size", "Exec Flag"]

logger = logging.getLogger(__name__)


# Shim Cache format used by Windows 5.2 and 6.0 (Server 2003 through Vista/Server 2008)
class CacheEntryNt5(object):
    def __init__(self, is_32_bit, data=None):
        self.w_length = None
        self.w_maximum_length = None
        self.offset = None
        self.dw_low_date_time = None
        self.dw_high_date_time = None
        self.dw_file_size_low = None
        self.dw_file_size_high = None
        self.is_32_bit = is_32_bit
        if data is not None:
            self.update(data)

    def update(self, data):
        if self.is_32_bit:
            entry = struct.unpack("<2H 3L 2L", data)
        else:
            entry = struct.unpack("<2H 4x Q 2L 2L", data)
        self.w_length = entry[0]
        self.w_maximum_length = entry[1]
        self.offset = entry[2]
        self.dw_low_date_time = entry[3]
        self.dw_high_date_time = entry[4]
        self.dw_file_size_low = entry[5]
        self.dw_file_size_high = entry[6]

    def size(self):

        if self.is_32_bit:
            return NT5_2_ENTRY_SIZE32
        else:
            return NT5_2_ENTRY_SIZE64


# Shim Cache format used by Windows 6.1 (Win7 through Server 2008 R2).
class CacheEntryNt6(object):
    def __init__(self, is_32_bit, data=None):
        self.w_length = None
        self.w_maximum_length = None
        self.offset = None
        self.dw_low_date_time = None
        self.dw_high_date_time = None
        self.file_flags = None
        self.flags = None
        self.blob_size = None
        self.blob_offset = None
        self.is_32_bit = is_32_bit
        if data is not None:
            self.update(data)

    def update(self, data):
        if self.is_32_bit:
            entry = struct.unpack("<2H 7L", data)
        else:
            entry = struct.unpack("<2H 4x Q 4L 2Q", data)
        self.w_length = entry[0]
        self.w_maximum_length = entry[1]
        self.offset = entry[2]
        self.dw_low_date_time = entry[3]
        self.dw_high_date_time = entry[4]
        self.file_flags = entry[5]
        self.flags = entry[6]
        self.blob_size = entry[7]
        self.blob_offset = entry[8]

    def size(self):
        if self.is_32_bit:
            return NT6_1_ENTRY_SIZE32
        else:
            return NT6_1_ENTRY_SIZE64


# Convert FILETIME to datetime.
# Based on http://code.activestate.com/recipes/511425-filetime-to-datetime/
def convert_filetime(dw_low_date_time, dw_high_date_time):
    try:
        date = datetime.datetime(1601, 1, 1, 0, 0, 0)
        temp_time = dw_high_date_time
        temp_time <<= 32
        temp_time |= dw_low_date_time
        return pytz.utc.localize(date + datetime.timedelta(microseconds=temp_time / 10))
    except OverflowError:
        return None


# Return a unique list while preserving ordering.
def unique_list(li):
    ret_list = []
    for entry in li:
        if entry not in ret_list:
            ret_list.append(entry)
    return ret_list


# Read the Shim Cache format, return a list of last modified dates/paths.
def get_shimcache_entries(cachebin, as_json=False):

    if len(cachebin) < 16:
        # Data size less than minimum header size.
        return None

    # Get the format type
    magic = struct.unpack("<L", cachebin[0:4])[0]

    # This is a Windows 2k3/Vista/2k8 Shim Cache format,
    if magic == CACHE_MAGIC_NT5_2:

        # Shim Cache types can come in 32-bit or 64-bit formats. We can
        # determine this because 64-bit entries are serialized with u_int64
        # pointers. This means that in a 64-bit entry, valid UNICODE_STRING
        # sizes are followed by a NULL DWORD. Check for this here.
        test_size = struct.unpack("<H", cachebin[8:10])[0]
        test_max_size = struct.unpack("<H", cachebin[10:12])[0]
        if (
            test_max_size - test_size == 2 and struct.unpack("<L", cachebin[12:16])[0]
        ) == 0:
            logger.debug("[+] Found 64bit Windows 2k3/Vista/2k8 Shim Cache data...")
            entry = CacheEntryNt5(False)
            yield from read_nt5_entries(cachebin, entry, as_json=as_json)

        # Otherwise it's 32-bit data.
        else:
            logger.debug("[+] Found 32bit Windows 2k3/Vista/2k8 Shim Cache data...")
            entry = CacheEntryNt5(True)
            yield from read_nt5_entries(cachebin, entry, as_json=as_json)

    # This is a Windows 7/2k8-R2 Shim Cache.
    elif magic == CACHE_MAGIC_NT6_1:
        test_size = struct.unpack(
            "<H", cachebin[CACHE_HEADER_SIZE_NT6_1: CACHE_HEADER_SIZE_NT6_1 + 2]
        )[0]
        test_max_size = struct.unpack(
            "<H", cachebin[CACHE_HEADER_SIZE_NT6_1 + 2: CACHE_HEADER_SIZE_NT6_1 + 4]
        )[0]

        # Shim Cache types can come in 32-bit or 64-bit formats.
        # We can determine this because 64-bit entries are serialized with
        # u_int64 pointers. This means that in a 64-bit entry, valid
        # UNICODE_STRING sizes are followed by a NULL DWORD. Check for this here.
        if (
            test_max_size - test_size == 2
            and struct.unpack(
                "<L",
                cachebin[CACHE_HEADER_SIZE_NT6_1 + 4: CACHE_HEADER_SIZE_NT6_1 + 8],
            )[0]
        ) == 0:
            logger.debug("[+] Found 64bit Windows 7/2k8-R2 Shim Cache data...")
            entry = CacheEntryNt6(False)
            yield from read_nt6_entries(cachebin, entry, as_json=as_json)
        else:
            logger.debug("[+] Found 32bit Windows 7/2k8-R2 Shim Cache data...")
            entry = CacheEntryNt6(True)
            yield from read_nt6_entries(cachebin, entry, as_json=as_json)

    # This is WinXP cache data
    elif magic == WINXP_MAGIC32:
        logger.debug("[+] Found 32bit Windows XP Shim Cache data...")
        yield from read_winxp_entries(cachebin, as_json=as_json)

    # Check the data set to see if it matches the Windows 8 format.
    elif (
        len(cachebin) > WIN8_STATS_SIZE
        and cachebin[WIN8_STATS_SIZE: WIN8_STATS_SIZE + 4] == WIN8_MAGIC
    ):
        logger.debug("[+] Found Windows 8/2k12 Apphelp Cache data...")
        yield from read_win8_entries(cachebin, WIN8_MAGIC, as_json=as_json)

    # Windows 8.1 will use a different magic dword, check for it
    elif (
        len(cachebin) > WIN8_STATS_SIZE
        and cachebin[WIN8_STATS_SIZE: WIN8_STATS_SIZE + 4] == WIN81_MAGIC
    ):
        logger.debug("[+] Found Windows 8.1 Apphelp Cache data...")
        yield from read_win8_entries(cachebin, WIN81_MAGIC, as_json=as_json)

    # Windows 10 will use a different magic dword, check for it
    elif (
        len(cachebin) > WIN10_STATS_SIZE
        and cachebin[WIN10_STATS_SIZE: WIN10_STATS_SIZE + 4] == WIN10_MAGIC
    ):
        logger.debug("[+] Found Windows 10 Apphelp Cache data...")
        yield from read_win10_entries(cachebin, WIN10_MAGIC, as_json=as_json)

    # Windows 10 creators update moved the damn magic 4 bytes forward...
    elif (
        len(cachebin) > WIN10_STATS_SIZE
        and cachebin[WIN10_STATS_SIZE + 4: WIN10_STATS_SIZE + 8] == WIN10_MAGIC
    ):
        logger.debug("[+] Found Windows 10 Apphelp Cache data... (creators update)")
        yield from read_win10_entries(
            cachebin, WIN10_MAGIC, creators_update=True, as_json=as_json
        )

    else:
        raise Exception(
            "Got an unrecognized magic value of 0x{:x}... bailing".format(magic)
        )


# Read Windows 8/2k12/8.1 Apphelp Cache entry formats.
def read_win8_entries(bin_data, ver_magic, as_json=False):

    entry_meta_len = 12

    # Skip past the stats in the header
    cache_data = bin_data[WIN8_STATS_SIZE:]

    data = sio.BytesIO(cache_data)
    while data.tell() < len(cache_data):
        header = data.read(entry_meta_len)
        # Read in the entry metadata
        # Note: the crc32 hash is of the cache entry data
        magic, crc32_hash, entry_len = struct.unpack("<4sLL", header)

        # Check the magic tag
        if magic != ver_magic:
            raise Exception(
                "Invalid version magic tag found: {}".format(
                    struct.unpack("I", magic)[0]
                )
            )

        entry_data = sio.BytesIO(data.read(entry_len))

        # Read the path length
        path_len = struct.unpack("<H", entry_data.read(2))[0]
        if path_len == 0:
            path = "None"
        else:
            path = entry_data.read(path_len).decode("utf-16le", "replace")

        # Check for package data
        package_len = struct.unpack("<H", entry_data.read(2))[0]
        if package_len > 0:
            # Just skip past the package data if present (for now)
            entry_data.seek(package_len, 1)

        # Read the remaining entry data
        flags, unk_1, low_datetime, high_datetime, unk_2 = struct.unpack(
            "<LLLLL", entry_data.read(20)
        )

        # Check the flag set in CSRSS
        if flags & CSRSS_FLAG:
            exec_flag = "True"
        else:
            exec_flag = "False"

        last_mod_date = convert_filetime(low_datetime, high_datetime)

        yield {
            "last_mod_date": last_mod_date.isoformat() if as_json else last_mod_date,
            "path": path,
            "exec_flag": exec_flag,
        }


# Read Windows 10 Apphelp Cache entry format
def read_win10_entries(bin_data, ver_magic, creators_update=False, as_json=False):

    entry_meta_len = 12

    # Skip past the stats in the header
    if creators_update:
        cache_data = bin_data[WIN10_STATS_SIZE + 4:]
    else:
        cache_data = bin_data[WIN10_STATS_SIZE:]

    data = sio.BytesIO(cache_data)
    while data.tell() < len(cache_data):
        header = data.read(entry_meta_len)
        # Read in the entry metadata
        # Note: the crc32 hash is of the cache entry data
        magic, crc32_hash, entry_len = struct.unpack("<4sLL", header)

        # Check the magic tag
        if magic != ver_magic:
            raise Exception(
                "Invalid version magic tag found: {}".format(
                    struct.unpack("<I", magic)[0]
                )
            )

        entry_data = sio.BytesIO(data.read(entry_len))

        # Read the path length
        path_len = struct.unpack("<H", entry_data.read(2))[0]
        if path_len == 0:
            path = "None"
        else:
            path = entry_data.read(path_len).decode("utf-16le", "replace")

        # Read the remaining entry data
        low_datetime, high_datetime = struct.unpack("<LL", entry_data.read(8))

        # Skip the unrecognized Microsoft App entry format for now
        if not (low_datetime + high_datetime):
            continue
        else:
            last_mod_date = convert_filetime(low_datetime, high_datetime)

        yield {
            "last_mod_date": last_mod_date.isoformat() if as_json else last_mod_date,
            "path": path,
        }


# Read Windows 2k3/Vista/2k8 Shim Cache entry formats.
def read_nt5_entries(bin_data, entry, as_json=False):
    contains_file_size = False
    entry_size = entry.size()

    num_entries = struct.unpack("<L", bin_data[4:8])[0]
    if num_entries == 0:
        return None

    # On Windows Server 2008/Vista, the filesize is swapped out of this
    # structure with two 4-byte flags. Check to see if any of the values in
    # "dwFileSizeLow" are larger than 2-bits. This indicates the entry contained file sizes.
    for offset in range(
        CACHE_HEADER_SIZE_NT5_2,
        (num_entries * entry_size) + CACHE_HEADER_SIZE_NT5_2,
        entry_size,
    ):

        entry.update(bin_data[offset: offset + entry_size])

        if entry.dw_file_size_low > 3:
            contains_file_size = True
            break

    # Now grab all the data in the value.
    for offset in range(
        CACHE_HEADER_SIZE_NT5_2,
        (num_entries * entry_size) + CACHE_HEADER_SIZE_NT5_2,
        entry_size,
    ):

        entry.update(bin_data[offset: offset + entry_size])

        last_mod_date = convert_filetime(
            entry.dw_low_date_time, entry.dw_high_date_time
        )
        path = bin_data[entry.offsets: entry.offset + entry.w_length].decode(
            "utf-16le", "replace"
        )

        # It contains file size data.
        exec_flag = None
        if contains_file_size:
            yield {
                "last_mod_date": (
                    last_mod_date.isoformat() if as_json else last_mod_date
                ),
                "path": path,
                "file_size": entry.dw_file_size_low,
            }

        # It contains flags.
        else:
            # Check the flag set in CSRSS
            if entry.dw_file_size_low & CSRSS_FLAG:
                exec_flag = "True"
            else:
                exec_flag = "False"

            yield {
                "last_mod_date": (
                    last_mod_date.isoformat() if as_json else last_mod_date
                ),
                "path": path,
                "exec_flag": exec_flag,
            }


# Read the Shim Cache Windows 7/2k8-R2 entry format,
# return a list of last modifed dates/paths.
def read_nt6_entries(bin_data, entry, as_json=False):
    entry_size = entry.size()
    num_entries = struct.unpack("<L", bin_data[4:8])[0]

    if num_entries == 0:
        return None

    # Walk each entry in the data structure.
    for offset in range(
        CACHE_HEADER_SIZE_NT6_1,
        num_entries * entry_size + CACHE_HEADER_SIZE_NT6_1,
        entry_size,
    ):

        entry.update(bin_data[offset: offset + entry_size])
        last_mod_date = convert_filetime(
            entry.dw_low_date_time, entry.dw_high_date_time
        )
        path = bin_data[entry.offset: entry.offset + entry.w_length].decode(
            "utf-16le", "replace"
        )

        # Test to see if the file may have been executed.
        if entry.file_flags & CSRSS_FLAG:
            exec_flag = "True"
        else:
            exec_flag = "False"

        yield {
            "last_mod_date": last_mod_date.isoformat() if as_json else last_mod_date,
            "path": path,
            "exec_flag": exec_flag,
        }


# Read the WinXP Shim Cache data. Some entries can be missing data but still
# contain useful information, so try to get as much as we can.
def read_winxp_entries(bin_data, as_json=False):

    num_entries = struct.unpack("<L", bin_data[8:12])[0]
    if num_entries == 0:
        return None

    for offset in range(
        WINXP_HEADER_SIZE32,
        (num_entries * WINXP_ENTRY_SIZE32) + WINXP_HEADER_SIZE32,
        WINXP_ENTRY_SIZE32,
    ):
        # No size values are included in these entries, so search for utf-16 terminator.
        path_len = bin_data[offset: offset + (MAX_PATH + 8)].find(b"\x00\x00")

        # if path is corrupt, procede to next entry.
        if path_len == 0:
            continue
        path = bin_data[offset: offset + path_len + 1].decode("utf-16le")

        if len(path) == 0:
            continue

        entry_data = offset + (MAX_PATH + 8)

        # Get last mod time.
        last_mod_time = struct.unpack("<2L", bin_data[entry_data: entry_data + 8])
        last_mod_time = convert_filetime(last_mod_time[0], last_mod_time[1])

        # Get last file size.
        file_size = struct.unpack("<2L", bin_data[entry_data + 8: entry_data + 16])[0]
        if file_size == 0:
            file_size = BAD_ENTRY_DATA

        # Get last update time.
        exec_time = struct.unpack("<2L", bin_data[entry_data + 16: entry_data + 24])
        exec_time = convert_filetime(exec_time[0], exec_time[1])

        yield {
            "last_mod_time": last_mod_time.isoformat() if as_json else last_mod_time,
            "exec_time": exec_time.isoformat() if as_json else exec_time,
            "path": path,
            "file_size": file_size,
        }


def parse_output(output):
    new_output_list = list()
    for row in output:
        exec_flag = False
        if row[4] == "True":
            exec_flag = True
        entry = {
            "last_mod_date": row[0],
            "last_update": row[1],
            "path": row[2],
            "file_size": row[3],
            "exec_flag": exec_flag,
        }
        new_output_list.append(entry)
    return new_output_list



import logging

from regipy import RegistryKeyNotFoundException, convert_wintime
from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)

UAC_KEY_PATH = r"\Microsoft\Windows\CurrentVersion\Policies\System"


class UACStatusPlugin(Plugin):
    NAME = "uac_plugin"
    DESCRIPTION = "Get the status of User Access Control"
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def run(self):
        """
        This plugin checks the following parameters:
        EnableLUA - Windows notifies the user when programs try to make changes to the computer
        EnableVirtualization - enables the redirection of legacy application File and Registry writes
                               that would normally fail as standard user to a user-writable data location.
        FilterAdministratorToken - If enabled approval is required when performing administrative tasks.
        ConsentPromptBehaviorAdmin - This option allows the Consent Admin to perform an
                                     operation that requires elevation without consent or credentials.
        ConsentPromptBehaviorUser - If enabled,  a standard user that needs to perform an operation that requires
                                    elevation of privilege will be prompted for an administrative user name and password
        """
        try:
            subkey = self.registry_hive.get_key(UAC_KEY_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(f"Could not find {self.NAME} subkey at {UAC_KEY_PATH}: {ex}")
            return None

        self.entries = {
            "last_write": convert_wintime(
                subkey.header.last_modified, as_json=self.as_json
            ),
            "enable_limited_user_accounts": subkey.get_value(
                "EnableLUA", as_json=self.as_json
            ),
            "enable_virtualization": subkey.get_value(
                "EnableVirtualization", as_json=self.as_json
            ),
            "filter_admin_token": subkey.get_value(
                "FilterAdministratorToken", as_json=self.as_json
            ),
            "consent_prompt_admin": subkey.get_value(
                "ConsentPromptBehaviorAdmin", as_json=self.as_json
            ),
            "consent_prompt_user": subkey.get_value(
                "ConsentPromptBehaviorUser", as_json=self.as_json
            ),
        }



# flake8: noqa
from .validation.validation_tests import shimcache_validation
from .validation.validation_tests import ntuser_persistence_validation
from .validation.validation_tests import software_persistence_validation
from .validation.validation_tests import ntuser_userassist_validation
from .validation.validation_tests import amcache_validation
from .validation.validation_tests import bam_validation
from .validation.validation_tests import word_wheel_query_ntuser_validation
from .validation.validation_tests import computer_name_plugin_validation
from .validation.validation_tests import uac_status_plugin_validation
from .validation.validation_tests import software_classes_installer_plugin_validation
from .validation.validation_tests import ntuser_classes_installer_plugin_validation
from .validation.validation_tests import ras_tracing_plugin_validation
from .validation.validation_tests import installed_programs_software_plugin_validation
from .validation.validation_tests import last_logon_plugin_validation
from .validation.validation_tests import typed_urls_plugin_validation
from .validation.validation_tests import profile_list_plugin_validation
from .validation.validation_tests import print_demon_plugin_validation
from .validation.validation_tests import services_plugin_validation
from .validation.validation_tests import local_sid_plugin_validation
from .validation.validation_tests import boot_key_plugin_validation
from .validation.validation_tests import host_domain_name_plugin_validation
from .validation.validation_tests import domain_sid_plugin_validation
from .validation.validation_tests import boot_entry_list_plugin_validation
from .validation.validation_tests import wdigest_plugin_validation
from .validation.validation_tests import winrar_plugin_validation
from .validation.validation_tests import network_drives_plugin_validation
from .validation.validation_tests import winscp_saved_sessions_plugin_validation
from .validation.validation_tests import usbstor_plugin_validation
from .validation.validation_tests import typed_paths_plugin_validation
from .validation.validation_tests import shell_bag_ntuser_plugin_validation
from .validation.validation_tests import shell_bag_usrclass_plugin_validation
from .validation.validation_tests import network_data_plugin_validation
from .validation.validation_tests import active_control_set_validation
from .validation.validation_tests import backuprestore_plugin_validation
from .validation.validation_tests import codepage_validation
from .validation.validation_tests import crash_dump_validation
from .validation.validation_tests import diag_sr_validation
from .validation.validation_tests import disable_last_access_validation
from .validation.validation_tests import disablesr_plugin_validation
from .validation.validation_tests import image_file_execution_options_validation
from .validation.validation_tests import installed_programs_ntuser_validation
from .validation.validation_tests import previous_winver_plugin_validation
from .validation.validation_tests import processor_architecture_validation
from .validation.validation_tests import routes_validation
from .validation.validation_tests import safeboot_configuration_validation
from .validation.validation_tests import shutdown_validation
from .validation.validation_tests import susclient_plugin_validation
from .validation.validation_tests import spp_clients_plugin_validation
from .validation.validation_tests import shutdown_validation
from .validation.validation_tests import terminal_services_history_validation
from .validation.validation_tests import timezone_data_validation
from .validation.validation_tests import timezone_data2_validation
from .validation.validation_tests import winrar_plugin_validation
from .validation.validation_tests import winver_plugin_validation
from .validation.validation_tests import wsl_plugin_validation


import logging

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime
from regipy.exceptions import RegistryKeyNotFoundException


logger = logging.getLogger(__name__)


LAST_ACCESS_PATH = r"Control\FileSystem"
crash_items = ("NtfsDisableLastAccessUpdate", "NtfsDisableLastAccessUpdate")
last_acc = {
    "80000000": "(User Managed, Updates Enabled)",
    "80000001": "(User Managed, Updates Disabled)",
    "80000002": "(System Managed, Updates Enabled)",
    "80000003": "(System Managed, Updates Disabled)",
}


class DisableLastAccessPlugin(Plugin):
    NAME = "disable_last_access"
    DESCRIPTION = "Get NTFSDisableLastAccessUpdate value"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def can_run(self):
        return self.registry_hive.hive_type == SYSTEM_HIVE_TYPE

    def run(self):
        self.entries = {}
        access_subkeys = self.registry_hive.get_control_sets(LAST_ACCESS_PATH)
        for access_subkey in access_subkeys:
            try:
                access = self.registry_hive.get_key(access_subkey)
            except RegistryKeyNotFoundException as ex:
                logger.error(
                    f"Could not find {self.NAME} subkey at {access_subkey}: {ex}"
                )
                continue
            self.entries[access_subkey] = {
                "last_write": convert_wintime(access.header.last_modified).isoformat()
            }
            for val in access.iter_values():
                if val.name in crash_items:
                    self.entries[access_subkey][val.name] = f"{val.value:0x}"
                    self.entries[access_subkey][f"{val.name}Str"] = last_acc.get(
                        f"{val.value:0x}", ""
                    )



import logging

from regipy import RegistryKeyNotFoundException, convert_wintime
from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)

CLASSES_INSTALLER_PATH = r"\Classes\Installer\Products"


class SoftwareClassesInstallerPlugin(Plugin):
    NAME = "software_classes_installer"
    DESCRIPTION = "List of installed software from SOFTWARE hive"
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def run(self):
        try:
            installer_subkey = self.registry_hive.get_key(CLASSES_INSTALLER_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(ex)
            return

        for entry in installer_subkey.iter_subkeys():
            identifier = entry.name
            timestamp = convert_wintime(
                entry.header.last_modified, as_json=self.as_json
            )
            product_name = entry.get_value("ProductName")
            self.entries.append(
                {
                    "identifier": identifier,
                    "timestamp": timestamp,
                    "product_name": product_name,
                    "is_hidden": product_name is None,
                }
            )
